import numpy as np
from sklearn.model_selection import train_test_split
import tensorflow as tf
import keras
import gen
from keras.optimizers import Adam, SGD
import matplotlib.pyplot as plt
from itertools import product
from utils import save_data, load_data
import os

os.environ["CUDA_VISIBLE_DEVICES"]= "1"


SUPERVISED_ONLY = "supervised_only"
WITH_MANIFOLD = "with-manifold"

ID = gen.FRANK
OPTIMIZER = gen.GD

acc = []
val_acc = []

documents = np.load("datasets/citeseer/words.npy")
labels = np.load("datasets/citeseer/labels.npy")
citations = np.load("datasets/citeseer/citations.npy")
citations[citations > 0] = 1
citations[citations < 0] = 0
num_documents = len(documents)
output_size = 6
labels = np.eye(output_size)[labels]
pairs_ids_xy = np.transpose(np.greater(citations, 0).nonzero())




epochs = 1000

sess = tf.Session()


class FeedForwardNN(keras.Sequential):

    def __init__(self, input_shape, output_size):
        super(FeedForwardNN, self).__init__()
        self.add(keras.layers.Dense(units=30, activation="relu", input_shape=input_shape, kernel_regularizer=tf.keras.regularizers.l2(0.001)))
        # self.add(keras.layers.Dense(units=20, activation="relu"))
        # self.add(keras.layers.Dense(units=100, activation="relu"))
        self.add(keras.layers.Dense(units=output_size, activation="softmax"))


np.random.seed(0)
train_ids = tf.placeholder(tf.int32, shape = [None])
test_ids = tf.placeholder(tf.int32, shape = [None])

P = tf.placeholder(tf.float32, shape=())
is_manifold = tf.placeholder(tf.bool, shape=())
LR = tf.placeholder(tf.float32, shape=())
rule_weight = tf.placeholder(tf.float32, shape=())

def logic(y_true, y_pred, ids_input):
    ids_input = tf.cast(ids_input, tf.int64)
    TNorm, Generator = gen.TGFactory.getTG(id=ID)
    g = Generator(p=P)
    print(g)
    logic = gen.Logic(g)
    KB = gen.Knowledge()

    kk = tf.gather(y_true, ids_input)
    tt = tf.gather(y_pred, ids_input)

    KB.add(tf.reduce_sum(logic.IMPLIES(kk, tt)))
    loss = KB.loss()
    return loss


def accuracy(y_true, y_pred):
    return tf.reduce_mean(tf.cast(tf.equal(tf.argmax(y_true, axis=1), tf.argmax(y_pred, axis=1)), tf.float32))


model = FeedForwardNN(output_size=output_size, input_shape=np.shape(documents)[1:])
y_pred = model(tf.constant(documents, dtype=tf.float32))

loss = logic(tf.cast(labels, tf.float32), y_pred, train_ids)
if OPTIMIZER == gen.ADAM:
    train_op = tf.train.AdamOptimizer(LR).minimize(loss)
else:
    train_op = tf.train.GradientDescentOptimizer(LR).minimize(loss)

y_true_train = tf.gather(labels, train_ids)
y_pred_train = tf.gather(y_pred, train_ids)

y_true_test = tf.gather(labels, test_ids)
y_pred_test = tf.gather(y_pred, test_ids)

train_acc = accuracy(y_true_train, y_pred_train)
test_acc = accuracy(y_true_test, y_pred_test)


def train(s, p, MODE, lr, w):


    TRID, TEID = train_test_split(np.arange(num_documents), test_size=s)

    tf.set_random_seed(0)
    np.random.seed(0)
    sess.run(tf.global_variables_initializer())
    val_acc = []
    acc = []
    feed_dict = {P:p, is_manifold: MODE==WITH_MANIFOLD, LR:lr, rule_weight:w, train_ids: TRID, test_ids:TEID}
    for e in range(epochs):

        sess.run(train_op, feed_dict=feed_dict)
        tra, tea = sess.run((train_acc, test_acc), feed_dict=feed_dict)
        print("Epoch {}/{}: --- Train Acc: {:.4f} --- Test Acc: {:.4f}".format(e,epochs,tra,tea))
        val_acc.append(tea)
        acc.append(tra)

    return acc, val_acc



if __name__ == "__main__":

    repetitions = 10
    SPLITS = [0.1]
    MODES = [SUPERVISED_ONLY]
    if ID == gen.SS:
        PS = [-1.5, -1, 0, 1, 1.5]
        if OPTIMIZER==gen.GD:
            LRS = [0.00001]
        else:
            LRS = [0.00001]
        # PS = [0]
    else:
        PS = [0.1, 1.,  100,  10000, float('inf')]
        if OPTIMIZER == gen.GD:
            LRS = [0.00001]
        else:
            LRS = [0.00001]

    w = 0

    exp_dict = {}

    for exp_id in product(SPLITS, MODES, PS, LRS):

        s, m, p, lr = exp_id

        AC = []
        VAL_ACC = []

        for d in range(repetitions):
            ac, val_ac = train(s, p, m, lr, w)
            AC.append(ac)
            VAL_ACC.append(val_ac)
        exp_dict[(s,m,p,lr)] = (np.mean(AC, axis=0), np.mean(VAL_ACC, axis=0))
        save_data(exp_dict, "exp_dict_{}_{}.pkl".format(ID, OPTIMIZER))



    exp_dict = load_data("exp_dict_{}_{}.pkl".format(ID, OPTIMIZER))
    to_plot = []
    # for key,value in exp_dict.items():
    #     if key[2] == 1.5:
    #         to_plot.append((key[3],value))

    # for k,v in to_plot:
    #     plt.plot(v[0], label="tr"+str(k))

    for k,v in exp_dict.items():
        plt.plot(v[1][:200], label="$\lambda = $"+str(k[2]))

    plt.legend(loc='lower right')
    plt.savefig("learning_dynamics_{}_{}.pdf".format(ID.lower(), OPTIMIZER.lower()))
    plt.clf()

    # for i in range(len(P)):
    #     plt.plot(val_acc[i], label="{}-{:.3f}".format(P[i], val_acc[i][-1]))
    #
    # plt.legend()
    # plt.title("{}".format(m))
    # plt.savefig("supervised_only_test.pdf".format(m))
    # plt.clf()
    #
    # val_acc = []
    # acc = []






