import tensorflow as tf
eps=1e-12


def r(x):
    return tf.where(x < eps, x + eps, x)


FRANK = "FRANK"
SS="SS"
PRODUCT = "PRODUCT"
ADAM = "Adam"
GD = "GD"

def log_b(x, b):
    numerator = tf.log(x + eps)
    denominator = tf.log(tf.constant(b, dtype=numerator.dtype) + eps)
    return numerator / denominator

class TNorm(object):
    def __init__(self):
        pass

class Generator(object):
    def __init__(self,p):
        pass


class ProductTNorm(TNorm):

    def __init__(self):
        super(ProductTNorm, self).__init__()

    def __call__(self, x,y):
        return x * y

class ProductGenerator(Generator):

    def __init__(self,p):
        super(ProductGenerator, self).__init__(p)

    def __call__(self, x):
        eps = 1e-7
        x = tf.clip_by_value(x, eps, 1. - eps)
        return - tf.log(x)


class FrankTNorm(TNorm):

    def __init__(self, p):
        super(FrankTNorm, self).__init__()
        self.p = p if p == "inf" else 1. * p

    def __call__(self, x, y):
        p = self.p
        if self.p == 0:
            return tf.minimum(x, y)
        elif self.p == 1:
            return x * y
        elif self.p == "inf":
            return tf.maximum(0., x + y - 1.)

        else:
            return log_b(1. + (tf.pow(p, x) - 1) * (tf.pow(p, y) - 1) / (p - 1), p)


class FrankGenerator(Generator):

    def __init__(self, p):
        super(FrankGenerator,self).__init__(p)
        self.p = p

    def __call__(self, x):
        def r(x):
            return tf.minimum(x + 1e-6, 1)
        p = self.p

        res = tf.case(
            pred_fn_pairs = [(tf.equal(p,1.), lambda:- tf.log(x +  1e-6)),
                             (tf.equal(p,float('inf')), lambda: 1 - x)],
            default= lambda: tf.log((p - 1.)/(tf.pow(p, r(x)) - 1)),
            exclusive = True)
        return res

class SSTNorm(TNorm):

    def __init__(self, p):
        super(SSTNorm, self).__init__()
        self.p = p if p == "inf" or p =="-inf" else 1. * p

    def __call__(self, x, y):
        def r(x):
            return tf.minimum(x + eps, 1)
        p = self.p
        if self.p == "-inf":
            return tf.minimum(x, y)
        elif p < 0:
            return tf.pow(tf.pow(r(x), p) + tf.pow(r(y), p) - 1, 1/p)
        elif self.p == 0:
            return x * y
        elif p > 0:
            return tf.pow(tf.maximum(0., tf.pow(r(x), p) + tf.pow(r(y), p) - 1), 1/p)
        elif self.p == "inf":
            return tf.maximum(0., x + y - 1.)

class SSGenerator(Generator):

    def __init__(self, p):
        super(SSGenerator,self).__init__(p)
        self.p = p

    def __call__(self, x):
        p = self.p
        return tf.cond(tf.equal(p,0),
                lambda: - tf.log(x + eps),
                lambda: (1 - tf.pow(x+eps, p)) / p)


class TGFactory(object):

    def __init__(self):
        pass
    @staticmethod
    def getTG(id):
        if id==FRANK:
            return FrankTNorm, FrankGenerator
        elif id == SS:
            return SSTNorm, SSGenerator
        elif id == PRODUCT:
            return ProductTNorm, ProductGenerator
        else:
            raise Exception("Unknown id for TNorm and Generato Family")




class Logic():

    def __init__(self, generator):
        self.g = generator

    def reduce_and(self, x, axis=1):

        return tf.reduce_sum(x, axis=axis)

    def and_n(self, x):

        return tf.add_n(x)

    def IMPLIES(self, x, y):

        return tf.where(x>y, self.g(y) - self.g(x), tf.zeros_like(x))

    def IFF(self, x, y):

        return self.IMPLIES(x,y) + self.IMPLIES(y,x)

    def FORALL(self, x, axis=0):

        return tf.reduce_sum(x, axis=axis)

    def reduce_or(self, x, axis = 1):

        return tf.reduce_min(x, axis= axis)

    def implies(self, x, y):

        return tf.where(x < y, y - x, tf.zeros_like(x))

    def or_n(self, x):

        x = tf.stack(x, axis=1)
        return self.reduce_or(x, axis=1)


class Knowledge():

    def __init__(self):
        self.rules = []

    def add(self, rule, weight=None):
        if weight is not None:
            self.rules.append(weight*rule)
        else:
            self.rules.append(rule)

    def loss(self):
        return tf.add_n(self.rules)