import numpy as np
from sklearn.model_selection import train_test_split
import tensorflow as tf
import keras
import gen
from keras.optimizers import Adam, SGD
import matplotlib.pyplot as plt
from itertools import product


SUPERVISED_ONLY = "supervised_only"
WITH_MANIFOLD = "with-manifold"

ID = gen.SS

acc = []
val_acc = []

documents = np.load("datasets/citeseer/words.npy")
labels = np.load("datasets/citeseer/labels.npy")
citations = np.load("datasets/citeseer/citations.npy")
citations[citations > 0] = 1
citations[citations < 0] = 0
num_documents = len(documents)
output_size = 6
labels = np.eye(output_size)[labels]
pairs_ids_xy = np.transpose(np.greater(citations, 0).nonzero())
repetitions = 10


epochs = 500

sess = tf.Session()


class FeedForwardNN(keras.Sequential):

    def __init__(self, input_shape, output_size):
        super(FeedForwardNN, self).__init__()
        self.add(keras.layers.Dense(units=100, activation="relu", input_shape=input_shape))
        self.add(keras.layers.Dense(units=100, activation="relu"))
        self.add(keras.layers.Dense(units=100, activation="relu"))
        self.add(keras.layers.Dense(units=output_size, activation="softmax"))


np.random.seed(0)
train_ids = tf.placeholder(tf.int32, shape = [None])
test_ids = tf.placeholder(tf.int32, shape = [None])

P = tf.placeholder(tf.float32, shape=())
is_manifold = tf.placeholder(tf.bool, shape=())
LR = tf.placeholder(tf.float32, shape=())
rule_weight = tf.placeholder(tf.float32, shape=())

def logic(y_true, y_pred, ids_input):
    ids_input = tf.cast(ids_input, tf.int64)
    TNorm, Generator = gen.TGFactory.getTG(id=gen.SS)
    g = Generator(p=P)
    logic = gen.Logic(g)
    KB = gen.Knowledge()

    #
    yy = g(y_pred)
    xy = tf.gather(yy, pairs_ids_xy, axis=0)  # shape(xy) = [num_citations, 2, num_outputs]
    impl = logic.implies(xy[:, 0, :], xy[:, 1, :])  #
    manifold_rule = tf.reduce_sum(logic.reduce_and(impl))
    KB.add(tf.cond(is_manifold, lambda: manifold_rule, lambda: tf.constant(0.)), weight=rule_weight)

    kk = tf.gather(y_true, ids_input)
    tt = tf.gather(y_pred, ids_input)

    KB.add(tf.reduce_sum(logic.IMPLIES(kk, tt)))
    loss = KB.loss()
    return loss


def accuracy(y_true, y_pred):
    return tf.reduce_mean(tf.cast(tf.equal(tf.argmax(y_true, axis=1), tf.argmax(y_pred, axis=1)), tf.float32))


model = FeedForwardNN(output_size=output_size, input_shape=np.shape(documents)[1:])
y_pred = model(tf.constant(documents, dtype=tf.float32))

loss = logic(tf.cast(labels, tf.float32), y_pred, train_ids)
train_op = tf.train.AdamOptimizer(LR).minimize(loss)

y_true_train = tf.gather(labels, train_ids)
y_pred_train = tf.gather(y_pred, train_ids)

y_true_test = tf.gather(labels, test_ids)
y_pred_test = tf.gather(y_pred, test_ids)

train_acc = accuracy(y_true_train, y_pred_train)
test_acc = accuracy(y_true_test, y_pred_test)


def train(s, p, MODE, lr, w):


    TRID, TEID = train_test_split(np.arange(num_documents), test_size=s)

    tf.set_random_seed(0)
    np.random.seed(0)
    sess.run(tf.global_variables_initializer())
    val_acc = []
    acc = []
    feed_dict = {P:p, is_manifold: MODE==WITH_MANIFOLD, LR:lr, rule_weight:w, train_ids: TRID, test_ids:TEID}
    for e in range(epochs):

        sess.run(train_op, feed_dict=feed_dict)
        tra, tea = sess.run((train_acc, test_acc), feed_dict=feed_dict)
        print("Epoch {}/{}: --- Train Acc: {:.4f} --- Test Acc: {:.4f}".format(e,epochs,tra,tea))
        val_acc.append(tea)
        acc.append(tra)

    return acc[-1], val_acc[-1]



if __name__ == "__main__":

    with open("log_diff_generators.csv", "w") as f:
        SPLITS = [0.1, 0.25, 0.5, 0.75, 0.9]
        LRS = [0.001, 0.0001, 0.00001]
        MODES = [SUPERVISED_ONLY, WITH_MANIFOLD]
        if ID == gen.SS:
            # PS = [-1.5, -1, 0, 1, 1.5]
            PS = [0]
        else:
            PS = [0]
        WEIGHTS = [0.1, 0.01, 0.006, 0.003, 0.001,0.0001]


        for exp_id in product(SPLITS, MODES, PS, LRS):

            s, m, p, lr = exp_id

            if m==WITH_MANIFOLD:
                WS = WEIGHTS
            else:
                WS = [0]


            for w in WS:
                AC = []
                VAL_AC = []
                for d in range(repetitions):
                    ac, val_ac = train(s, p, m, lr, w)
                    AC.append(ac)
                    VAL_AC.append(val_ac)
                lll = "\t".join([str(i) for i in [s,p,m,lr,w,np.mean(AC), np.mean(VAL_AC), np.std(VAL_AC)]])
                f.write(lll+"\n")

        # np.save("val_acc_{}.npy".format(m), np.array(val_acc))
        # np.save("acc_{}.npy".format(m), np.array(acc))
        #
        # val_acc = np.load("val_acc_{}.npy".format(m))
        # acc = np.load("acc_{}.npy".format(m))
        #
        # for i in range(len(P)):
        #     plt.plot(acc[i], label="{}-{:.3f}".format(P[i], acc[i][-1]))
        #
        # plt.legend()
        # plt.title("{}".format(m))
        # plt.savefig("citeseer_manifold_train_{}.pdf".format(m))
        # plt.clf()
        #
        # for i in range(len(P)):
        #     plt.plot(val_acc[i], label="{}-{:.3f}".format(P[i], val_acc[i][-1]))
        #
        # plt.legend()
        # plt.title("{}".format(m))
        # plt.savefig("citeseer_manifold_test_{}.pdf".format(m))
        # plt.clf()
        #
        # val_acc = []
        # acc = []




